var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _GossipLogService_instances, _GossipLogService_registrarId, _GossipLogService_pubsub, _GossipLogService_started, _GossipLogService_push, _GossipLogService_sync;
import { TypedEventEmitter, TopicValidatorResult, } from "@libp2p/interface";
import { GossipSub } from "@chainsafe/libp2p-gossipsub";
import { logger } from "@libp2p/logger";
import * as lp from "it-length-prefixed";
import { pipe } from "it-pipe";
import { pushable } from "it-pushable";
import { equals } from "uint8arrays";
import { anySignal } from "any-signal";
import PQueue from "p-queue";
import { assert } from "@canvas-js/utils";
import { Event } from "#protocols/events";
import { DEFAULT_PROTOCOL_SELECT_TIMEOUT, MAX_INBOUND_STREAMS, MAX_OUTBOUND_STREAMS, NEGOTIATE_FULLY, PUSH_RETRY_INTERVAL, PUSH_RETRY_LIMIT, SYNC_RETRY_INTERVAL, SYNC_RETRY_LIMIT, SYNC_TIMEOUT, second, } from "../constants.js";
import { decodeId, encodeId } from "../ids.js";
import { Client, decodeRequests, encodeResponses } from "../sync/index.js";
import { DelayableController, SyncTimeoutError, wait } from "../utils.js";
import { Server } from "../sync/server.js";
export const getSyncProtocol = (topic) => `/canvas/v1/${topic}/sync`;
export const getPushProtocol = (topic) => `/canvas/v1/${topic}/push`;
export class GossipLogService extends TypedEventEmitter {
    static extractGossipSub(components) {
        const { pubsub } = components;
        assert(pubsub !== undefined, "expected pubsub !== undefined");
        assert(pubsub instanceof GossipSub, "expected pubsub instanceof GossipSub");
        return pubsub;
    }
    constructor(components, messageLog, init) {
        super();
        _GossipLogService_instances.add(this);
        this.components = components;
        this.messageLog = messageLog;
        this.syncProtocol = getSyncProtocol(this.messageLog.topic);
        this.pushProtocol = getPushProtocol(this.messageLog.topic);
        this.syncQueue = new PQueue({ concurrency: 1 });
        this.syncQueuePeers = new Set();
        this.pushQueue = new PQueue({ concurrency: 16 });
        this.topologyPeers = new Set();
        this.controller = new AbortController();
        _GossipLogService_registrarId.set(this, null);
        _GossipLogService_pubsub.set(this, void 0);
        _GossipLogService_started.set(this, false);
        this.forwardEvent = (event) => this.safeDispatchEvent(event.type, event);
        this.handleMessage = ({ detail: { msgId, propagationSource, msg } }) => {
            if (msg.topic !== this.messageLog.topic) {
                return;
            }
            this.log("received gossipsub message %s via %p", msgId, propagationSource);
            const sourceId = propagationSource.toString();
            let event;
            try {
                event = Event.decode(msg.data);
            }
            catch (err) {
                console.error("error decoding gossipsub message", err);
                this.log.error("error decoding gossipsub message: %O", err);
                __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Reject);
                return;
            }
            if (event.insert !== undefined) {
                const { key, value } = event.insert;
                this.handleInsert([key, value]).then((result) => {
                    __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").reportMessageValidationResult(msgId, sourceId, result);
                    if (result === TopicValidatorResult.Ignore) {
                        this.scheduleSync(propagationSource);
                    }
                }, (err) => {
                    this.log.error("error handling gossipsub message: %O", err);
                    __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Reject);
                });
            }
            else {
                console.error("IGNORING MESSAGE");
                __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Ignore);
            }
        };
        this.handleIncomingPush = async ({ connection, stream }) => {
            const peerId = connection.remotePeer;
            this.log("opened incoming push stream %s from peer %p", stream.id, peerId);
            await pipe(stream.source, lp.decode, async (msgs) => {
                try {
                    for await (const msg of msgs) {
                        const event = Event.decode(msg.subarray());
                        if (event.insert !== undefined) {
                            // ...
                        }
                        else if (event.update !== undefined) {
                            this.handleUpdate(connection.remotePeer, event.update.heads);
                        }
                    }
                }
                catch (err) {
                    stream.abort(err instanceof Error ? err : new Error(`${err}`));
                }
                // try {
                // 	const { value: msg, done } = await msgs.next()
                // 	assert(done === false && msg !== undefined, "expected done === false && msg !== undefined")
                // 	const heads = cbor.decode<Uint8Array[]>(msg.subarray())
                // 	await msgs.next().then((result) => assert(result.done, "expected result.done"))
                // 	this.handleUpdate(connection.remotePeer, heads)
                // } catch (err) {
                // 	stream.close()
                // }
            });
            this.log("closed incoming push stream %s from peer %p", stream.id, peerId);
        };
        this.handleIncomingSync = async ({ connection, stream }) => {
            const peerId = connection.remotePeer;
            this.log("opened incoming sync stream %s from peer %p", stream.id, peerId);
            const timeoutController = new DelayableController(3 * second);
            const signal = anySignal([this.controller.signal, timeoutController.signal]);
            signal.addEventListener("abort", (err) => {
                if (stream.status === "open") {
                    stream.abort(new Error("TIMEOUT"));
                }
            });
            try {
                await this.messageLog.serve(async (txn) => {
                    const server = new Server(txn);
                    await pipe(stream.source, lp.decode, decodeRequests, async function* (reqs) {
                        for await (const req of reqs) {
                            timeoutController.delay();
                            yield req;
                        }
                    }, (reqs) => server.handle(reqs), encodeResponses, lp.encode, stream.sink);
                });
                this.log("closed incoming stream %s from peer %p", stream.id, peerId);
            }
            catch (err) {
                if (err instanceof Error && err.message === "TIMEOUT") {
                    this.log.error("timed out incoming stream %s from peer %p", stream.id, peerId);
                    stream.abort(err);
                }
                else if (err instanceof Error) {
                    this.log.error("aborting incoming stream %s from peer %p: %O", stream.id, peerId, err);
                    stream.abort(err);
                }
                else {
                    this.log.error("aborting incoming stream %s from peer %p: %O", stream.id, peerId, err);
                    stream.abort(new Error("internal error"));
                }
            }
            finally {
                signal.clear();
            }
        };
        this.log = logger(`canvas:gossiplog:[${this.messageLog.topic}]:service`);
        __classPrivateFieldSet(this, _GossipLogService_pubsub, GossipLogService.extractGossipSub(components), "f");
        this.maxInboundStreams = init.maxInboundStreams ?? MAX_INBOUND_STREAMS;
        this.maxOutboundStreams = init.maxOutboundStreams ?? MAX_OUTBOUND_STREAMS;
    }
    isStarted() {
        return __classPrivateFieldGet(this, _GossipLogService_started, "f");
    }
    async start() {
        this.log("start");
        this.messageLog.addEventListener("sync", this.forwardEvent);
        this.messageLog.addEventListener("commit", this.forwardEvent);
        this.messageLog.addEventListener("message", this.forwardEvent);
        __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").addEventListener("gossipsub:message", this.handleMessage);
        await this.components.registrar.handle(this.pushProtocol, this.handleIncomingPush, {
            maxInboundStreams: this.maxInboundStreams,
            maxOutboundStreams: this.maxOutboundStreams,
        });
        await this.components.registrar.handle(this.syncProtocol, this.handleIncomingSync, {
            maxInboundStreams: this.maxInboundStreams,
            maxOutboundStreams: this.maxOutboundStreams,
        });
        __classPrivateFieldSet(this, _GossipLogService_registrarId, await this.components.registrar.register(this.syncProtocol, {
            notifyOnTransient: false,
            onConnect: async (peerId, connection) => {
                this.topologyPeers.add(peerId.toString());
                this.log("connected to peer %p", peerId);
                // having one peer wait an initial randomized interval
                // reduces the likelihood of deadlock to near-zero,
                // but it could still happen.
                if (connection.direction === "inbound") {
                    const interval = second + Math.floor(Math.random() * SYNC_RETRY_INTERVAL);
                    this.log("waiting an initial %dms", interval);
                    await this.wait(interval);
                }
                this.schedulePush(peerId);
            },
            onDisconnect: (peerId) => {
                this.log("disconnected from %p", peerId);
                this.topologyPeers.delete(peerId.toString());
            },
        }), "f");
        __classPrivateFieldSet(this, _GossipLogService_started, true, "f");
    }
    async afterStart() {
        this.log("afterStart");
        __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").subscribe(this.messageLog.topic);
    }
    async beforeStop() {
        this.log("beforeStop");
        __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").unsubscribe(this.messageLog.topic);
    }
    async stop() {
        this.log("stop");
        __classPrivateFieldSet(this, _GossipLogService_started, false, "f");
        __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").removeEventListener("gossipsub:message", this.handleMessage);
        this.messageLog.removeEventListener("sync", this.forwardEvent);
        this.messageLog.removeEventListener("commit", this.forwardEvent);
        this.messageLog.removeEventListener("message", this.forwardEvent);
        await this.components.registrar.unhandle(this.syncProtocol);
        if (__classPrivateFieldGet(this, _GossipLogService_registrarId, "f") !== null) {
            this.components.registrar.unregister(__classPrivateFieldGet(this, _GossipLogService_registrarId, "f"));
            __classPrivateFieldSet(this, _GossipLogService_registrarId, null, "f");
        }
    }
    async append(payload, options = {}) {
        const signedMessage = await this.messageLog.append(payload, options);
        const recipients = this.publish(signedMessage);
        return {
            id: signedMessage.id,
            signature: signedMessage.signature,
            message: signedMessage.message,
            recipients,
        };
    }
    async insert(signature, message) {
        assert(message.topic === this.messageLog.topic, "wrong topic");
        const signedMessage = this.messageLog.encode(signature, message);
        this.messageLog.insert(signedMessage);
        const recipients = this.publish(signedMessage);
        return { id: signedMessage.id, recipients };
    }
    async publish(signedMessage) {
        if (!__classPrivateFieldGet(this, _GossipLogService_pubsub, "f").isStarted()) {
            return [];
        }
        const data = Event.encode({ insert: { key: signedMessage.key, value: signedMessage.value } });
        return __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").publish(this.messageLog.topic, data).then(({ recipients }) => {
            this.log("published message %s to %d recipients %O", signedMessage.id, recipients.length, recipients);
            return recipients;
        }, (err) => {
            this.log.error("failed to publish event: %O", err);
            return [];
        });
    }
    async handleInsert([key, value]) {
        const signedMessage = this.messageLog.decode(value);
        assert(equals(key, signedMessage.key), "invalid key");
        await this.messageLog.verifySignature(signedMessage.signature, signedMessage.message);
        await this.messageLog.insert(signedMessage);
        return TopicValidatorResult.Accept;
        // return await this.messageLog.write(async (txn) => {
        // 	const missingParents = await this.messageLog.getMissingParents(txn, message.parents)
        // 	if (missingParents.size !== 0) {
        // 		this.log("missing %d/%d parents", missingParents.size, message.parents.length)
        // 		return TopicValidatorResult.Ignore
        // 	}
        // 	try {
        // 		await this.messageLog.apply(txn, id, signature, message, [key, value])
        // 		return TopicValidatorResult.Accept
        // 	} catch (err) {
        // 		return TopicValidatorResult.Reject
        // 	}
        // })
    }
    async handleUpdate(propagationSource, heads) {
        this.log("handling update: %o", heads.map(decodeId));
        const result = await this.messageLog.tree.read((txn) => {
            const missingParents = new Set();
            for (const key of heads) {
                const leaf = txn.getNode(0, key);
                if (leaf === null) {
                    missingParents.add(decodeId(key));
                }
            }
            if (missingParents.size === 0) {
                return TopicValidatorResult.Accept;
            }
            else {
                return TopicValidatorResult.Ignore;
            }
        });
        // Need to sync
        if (result === TopicValidatorResult.Ignore) {
            this.scheduleSync(propagationSource);
        }
        return result;
    }
    schedulePush(peerId) {
        this.log("scheduling push to %p", peerId);
        this.pushQueue
            .add(() => this.push(peerId))
            .catch((err) => {
            this.log.error("push failed: %O", err);
        });
    }
    async push(peerId) {
        this.log("initiating push to %p", peerId);
        const connection = this.components.connectionManager
            .getConnections(peerId)
            .find((connection) => connection.transient === false);
        if (connection === undefined) {
            this.log("no longer connected to %p", peerId);
            return;
        }
        await this.retry("push", () => __classPrivateFieldGet(this, _GossipLogService_instances, "m", _GossipLogService_push).call(this, connection), {
            interval: PUSH_RETRY_INTERVAL,
            limit: PUSH_RETRY_LIMIT,
        });
    }
    scheduleSync(peerId) {
        const id = peerId.toString();
        if (this.syncQueuePeers.has(id)) {
            this.log("already queued sync with %p", peerId);
            return;
        }
        this.log("scheduling sync with %p", peerId);
        this.syncQueuePeers.add(id);
        this.syncQueue
            .add(() => this.sync(peerId))
            .catch((err) => this.log.error("sync failed: %O", err))
            .finally(() => this.syncQueuePeers.delete(id));
    }
    async sync(peerId) {
        this.log("initiating sync with %p", peerId);
        const connection = this.components.connectionManager
            .getConnections(peerId)
            .find((connection) => connection.transient === false);
        if (connection === undefined) {
            this.log("no longer connected to %p", peerId);
            return;
        }
        await this.retry("sync", () => __classPrivateFieldGet(this, _GossipLogService_instances, "m", _GossipLogService_sync).call(this, connection), {
            interval: SYNC_RETRY_INTERVAL,
            limit: SYNC_RETRY_LIMIT,
        });
    }
    async retry(name, callback, options) {
        for (let n = 0; n < options.limit; n++) {
            try {
                await callback();
                return;
            }
            catch (err) {
                this.log.error("%s failed: %O", name, err);
                if (this.controller.signal.aborted) {
                    break;
                }
                else {
                    const interval = Math.floor(Math.random() * options.interval);
                    this.log("waiting %dms before trying %s again (%d/%d)", interval, name, n + 1, options.limit);
                    await this.wait(interval);
                    continue;
                }
            }
        }
        throw new Error("exceeded sync retry limit");
    }
    async wait(interval) {
        await wait(interval, { signal: this.controller.signal });
    }
}
_GossipLogService_registrarId = new WeakMap(), _GossipLogService_pubsub = new WeakMap(), _GossipLogService_started = new WeakMap(), _GossipLogService_instances = new WeakSet(), _GossipLogService_push = async function _GossipLogService_push(connection) {
    const peerId = connection.remotePeer;
    const [_, heads] = await this.messageLog.getClock();
    this.log("pushing heads %o to peer %p", heads, peerId);
    const protocolSelectSignal = AbortSignal.timeout(DEFAULT_PROTOCOL_SELECT_TIMEOUT);
    const stream = await connection
        .newStream(this.pushProtocol, { negotiateFully: NEGOTIATE_FULLY, signal: protocolSelectSignal })
        .catch((err) => {
        this.log.error("failed to open outgoing push stream: %O", err);
        throw err;
    });
    this.log("opened outgoing push stream %s to peer %p", stream.id, peerId);
    try {
        const source = pushable();
        const data = Event.encode({ update: { heads: heads.map(encodeId) } });
        await Promise.all([pipe(source, lp.encode, stream.sink), source.push(data).end().onEmpty()]);
        await stream.close();
        this.log("closed outgoing push stream %s to peer %p", stream.id, peerId);
    }
    catch (err) {
        this.log.error("error sending push: %o", err);
        stream.abort(err instanceof Error ? err : new Error(`${err}`));
    }
}, _GossipLogService_sync = async function _GossipLogService_sync(connection) {
    if (connection.status !== "open") {
        throw new Error("connection closed");
    }
    const peerId = connection.remotePeer;
    const sourceId = peerId.toString();
    const protocolSelectSignal = AbortSignal.timeout(DEFAULT_PROTOCOL_SELECT_TIMEOUT);
    const stream = await connection
        .newStream(this.syncProtocol, { negotiateFully: NEGOTIATE_FULLY, signal: protocolSelectSignal })
        .catch((err) => {
        this.log.error("failed to open outgoing sync stream: %O", err);
        throw err;
    });
    this.log("opened outgoing sync stream %s to peer %p", stream.id, peerId);
    const timeoutController = new DelayableController(SYNC_TIMEOUT);
    const signal = anySignal([this.controller.signal, timeoutController.signal]);
    signal.addEventListener("abort", (err) => {
        if (stream.status === "open") {
            stream.abort(new SyncTimeoutError());
        }
    });
    this.log("starting sync with peer %p", peerId);
    const client = new Client(stream);
    let messageCount = 0;
    try {
        const result = await this.messageLog.sync(client, { sourceId });
        messageCount = result.messageCount;
    }
    finally {
        signal.clear();
        client.end();
        this.log("closed outgoing stream %s to peer %p", stream.id, peerId);
    }
    if (messageCount !== 0) {
        for (const peer of __classPrivateFieldGet(this, _GossipLogService_pubsub, "f").getSubscribers(this.messageLog.topic)) {
            this.schedulePush(peer);
        }
    }
};
export const gossiplog = (messageLog, init = {}) => (components) => new GossipLogService(components, messageLog, init);
